import concurrent
import json
import tempfile
from pathlib import Path

import aiohttp
import asyncio
import os

import cv2
from loguru import logger
from mhdwriter.args import WriteArgs
from mhdwriter.writer import write_mhd_raw
from tqdm import tqdm


def handle_cft_download(url: str):
    logger.info(f"Downloading cft dataset from '{url}'")
    token_url = parse_token(url)
    json_result = asyncio.run(download_single(token_url))
    result_dict = json.loads(json_result)
    write_args = WriteArgs(
        fov=result_dict["fov"],
        protocol=result_dict["protocol"],
        study_description=result_dict["study_description"],
        series_description=result_dict["series_description"],
        downsample_factor=int(result_dict["downsample_factor"]),
        date_time=result_dict["study_date"],
        skip_files=False,
    )
    if len(result_dict["urls"]) == 0:
        logger.error(f"No files found for study.")
        return

    with tempfile.TemporaryDirectory() as temp_dir:
        logger.info(f"Found {len(result_dict['urls'])} files to download")
        logger.info(f"Downloading to temp dir {temp_dir}")
        file_len = 0
        retry_count = 3
        file_list = []
        while len(file_list) < len(result_dict["urls"]) and retry_count > 0:
            asyncio.run(download_urls(result_dict["urls"], temp_dir))
            file_list = list(Path(temp_dir).glob("*"))
            retry_count -= 1

        if len(file_list) != len(result_dict["urls"]):
            logger.error(f"Failed to download all files. Only got {file_len}/{len(result_dict['urls'])}")
            return
        if str(file_list[0]).endswith(".jp2"):
            logger.info(f"Found jp2 files to extract")
            with tqdm(total=len(file_list)) as progress_bar:
                progress_bar.set_description(f"Extracting jp2 files")
                with concurrent.futures.ProcessPoolExecutor() as executor:
                    futures = [executor.submit(convert_jp2, str(task)) for task in file_list]
                    for _ in concurrent.futures.as_completed(futures):
                        progress_bar.update(1)

        curr_dir = os.getcwd()
        logger.info(f"Writing raw file")

        write_mhd_raw(
            Path(temp_dir),
            write_args,
            Path(curr_dir)
        )
        logger.info(f"Wrote file to output path: {curr_dir}")


def parse_token(url_string: str):
    logger.info(f"Downloading cft dataset from '{url_string}'")
    token_args = url_string.split("_")
    base_url = asyncio.run(expand_url(token_args[0]))
    token_url = f"{base_url}/exports/{_expand_uid(token_args[1])}/{_expand_uid(token_args[2])}"
    return token_url


def _expand_uid(suid):
    alphabet = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    alphabet_hash = {char: index for index, char in enumerate(alphabet)}
    base = len(alphabet)
    num = 0
    for i, char in enumerate(reversed(suid)):
        num += alphabet_hash[char] * (base ** i)
    hex_str = format(num, '032x')
    return '-'.join([hex_str[0:8], hex_str[8:12], hex_str[12:16], hex_str[16:20], hex_str[20:32]])


async def download_file(session, url, destination, progress_bar, retries=3):
    progress_bar.set_description(f"Downloading {os.path.basename(destination)}")
    try:
        async with session.get(url) as response:
            if response.status == 200:
                with open(destination, 'wb') as f:
                    while True:
                        chunk = await response.content.read(1024)
                        if not chunk:
                            break
                        f.write(chunk)
                    progress_bar.update(1)  # Update the progress bar
    except aiohttp.ClientError as e:
        if retries > 0:
            await asyncio.sleep(1)  # Wait a bit before retrying
            await download_file(session, url, destination, progress_bar, retries - 1)
        else:
            raise e


async def download_urls(urls, dest_folder):
    async with aiohttp.ClientSession() as session:
        with tqdm(total=len(urls)) as progress_bar:
            tasks = []
            for url in urls:
                clean_url = url.split('?')[0]
                filename = os.path.basename(clean_url)
                destination = os.path.join(dest_folder, filename)
                task = asyncio.create_task(download_file(session, url, destination, progress_bar))
                tasks.append(task)

            # Await all tasks
            await asyncio.gather(*tasks)


async def download_single(url):
    async with aiohttp.ClientSession() as session:
        return await fetch(session, url)


async def fetch(session, url):
    async with session.get(url) as response:
        return await response.text()


async def expand_url(url):
    async with aiohttp.ClientSession() as session:
        async with session.get("http://tinyurl.com/" + url) as response:
            return response.url


def convert_jp2(jp2_filename: str):
    image = cv2.imread(jp2_filename, cv2.IMREAD_UNCHANGED)
    if image is None:
        raise ValueError("Could not load the JP2 image. Make sure the file exists and your OpenCV installation "
                         "supports JPEG 2000.")
    tiff_filename = jp2_filename.replace(".jp2", ".tiff")
    cv2.imwrite(tiff_filename, image)
    os.remove(jp2_filename)



import json
from dataclasses import dataclass, replace
from datetime import date, datetime, time
from pathlib import Path
from typing import Dict, Optional, List

from loguru import logger
from mhdwriter.args import WriteArgs, WriteType
from mhdwriter.writer import write_mhd_raw



@dataclass
class ImportFilter:
    app: str
    roi: Optional[str]
    exposure: List[int]
    output_dir: Optional[str] = None
    include_rgb: bool = True
    downsample_factor: int = 0
    write_type: WriteType = WriteType.RAW


def run_import(cft_path: Path, import_filter: ImportFilter) -> None:
    if cft_path.suffix == ".json" and cft_path.name.startswith("study_detail"):
        convert_local_dataset(cft_path, import_filter)
    else:
        logger.error("Invalid CFT path. Please provide a url,.cft, or study_detail.json file")


def convert_local_dataset(cft_path: Path, import_filter: ImportFilter) -> None:
    logger.info(f"Converting local cft dataset from '{cft_path}' to mhd/raw")
    json_data = parse_json_file(cft_path)

    if json_data is None or "study_details" not in json_data:
        logger.error("Invalid study_detail.json file")
        return

    roi_dir = cft_path.parent.joinpath("rois")

    if not roi_dir.exists():
        logger.error("Missing rois directory")
        return

    study_details = json_data["study_details"]
    base_out_path = cft_path.parent.joinpath("mhds")
    base_out_path.mkdir(exist_ok=True, parents=True)

    study_date_str = study_details.get("study_date", "20000101")
    study_time_str = study_details.get("study_time", "000000")
    study_date_time = parse_date_time(study_date_str, study_time_str)

    if study_date_time is None:
        raise ValueError("Unable to parse date time from study_details file.")

    base_args = WriteArgs(
        fov=study_details["fov"],
        protocol=study_details["protocol"],
        study_description=study_details["study"],
        series_description=study_details["study"],
        downsample_factor=import_filter.downsample_factor,
        date_time=study_date_time,
        write_type=import_filter.write_type,
    )

    for roi_path in roi_dir.iterdir():
        if not roi_path.is_dir():
            continue
        roi_name = roi_path.name
        if import_filter.roi is not None and roi_name != import_filter.roi:
            continue
        for roi_app_path in roi_path.iterdir():
            if roi_app_path.name == "rgb":
                logger.info("Generating RGB mhd/raw for " + str(roi_app_path))
                if import_filter.downsample_factor > 0:
                    series_desc = f"{study_details['study']}_{roi_name}_RGB_d{import_filter.downsample_factor}"
                else:
                    series_desc = f"{study_details['study']}_{roi_name}_RGB"
                write_mhd_raw(
                    roi_app_path,
                    replace(base_args, is_rgb=True, series_description=series_desc),
                    base_out_path
                )
            elif roi_app_path.is_dir() and roi_app_path.name.startswith("ex"):
                for exp in import_filter.exposure:
                    if import_filter.downsample_factor > 0:
                        series_desc = f"{study_details['study']}_{roi_name}_{roi_app_path.name}_{exp}_d{import_filter.downsample_factor}"
                    else:
                        series_desc = f"{study_details['study']}_{roi_name}_{roi_app_path.name}_{exp}"
                    exp_dir = roi_app_path.joinpath(f"{exp}")
                    if not exp_dir.is_dir():
                        logger.error(f"Missing exposure directory {exp_dir}")
                        continue
                    write_mhd_raw(
                        exp_dir,
                        replace(base_args, is_rgb=False, series_description=series_desc),
                        base_out_path
                    )


def parse_json_file(file_path: Path) -> Optional[Dict]:
    try:
        with file_path.open('r') as file:
            return json.load(file)
    except Exception as e:
        logger.error(f"An error occurred while reading the file: {e}")
        return None


def parse_date_time(study_date_str: str, study_time_str: str) -> Optional[datetime]:
    try:
        study_date = date(
            int(study_date_str[:4]),
            int(study_date_str[4:6]),
            int(study_date_str[6:]),
        )
    except ValueError as e:
        print("Unable to read date str " + study_date_str)
        return
    try:
        study_time = time(
            int(study_time_str[0:2]),
            int(study_time_str[2:4]),
            int(study_time_str[4:6]),
        )
    except ValueError as e:
        print("Unable to read time str " + study_time_str)
        return
    return datetime.combine(study_date, study_time)



from pathlib import Path
from typing import Optional, List

import click
from loguru import logger

from xerra.importer import run_import, ImportFilter

from xerra.download import handle_cft_download

from mhdwriter.writer import WriteType


@click.group()
def cli():
    pass


@click.command()
@click.argument("path")
@click.option("--app", required=False, help="Filter based on ex/em")
@click.option("--roi", required=False, help="Specific ROI to convert to mhd")
@click.option("--output-path", required=False,help="Output directory for mhd")
@click.option("--downsampled", default=0, help="Downsampled factor (1=1x,2=2x,etx...)")
@click.option("--stacked", is_flag=True, default=False,help="Output jpeg stack RGB")
@click.option(
    "--exposure",
    "-e",
    multiple=True,
    help="exposures to include",
    default=[],
)
def generate_mhd(
        path: str,
        app: Optional[str] = None,
        roi: Optional[str] = None,
        exposure: List[int] = [],
        output_path: Optional[str] = None,
        downsampled: int = 0,
        stacked: bool = False,
):
    input_filter = ImportFilter(
        app=app,
        roi=roi,
        exposure=exposure,
        output_dir=output_path,
        downsample_factor=downsampled,
    )
    if stacked:
        input_filter.write_type = WriteType.NONE
    path = Path(path).resolve()
    if path.exists():
        run_import(path,input_filter)
    else:
        logger.error(f"Missing input path {path}")

@click.command()
@click.argument("path")
def download(path: str):
    handle_cft_download(path)



def main():
    cli.add_command(download)
    cli.add_command(generate_mhd)
    cli()


if __name__ == "__main__":
    main()
